#include "ExceptionHandler.hpp"
#include <ostream>

namespace alib {

std::vector < std::function < int ( std::ostream &, unsigned ) > > & ExceptionHandler::handlers ( ) {
	static std::vector < std::function < int ( std::ostream &, unsigned ) > > res = {
		[ ] ( std::ostream & out, unsigned level ) {
			try {
				throw;
			} catch ( ... ) {
				out << level << ' ' << "unexpected exception." << std::endl;
				return 127;
			}
		}, [ ] ( std::ostream & out, unsigned level ) {
			try {
				throw;
			} catch ( const std::exception & e ) {
				out << level << ' ' << "standard exception: " << e.what() << std::endl;
				return rethrow_if_nested ( out, e, level, 3 );
			}
		}
	};
	return res;
}

int ExceptionHandler::rethrow_if_nested ( std::ostream & out, const std::exception & e, unsigned level, int result ) {
	try {
		std::rethrow_if_nested ( e );
		return result;
	} catch ( ... ) {
		return handlerRec ( out, level + 1, handlers ( ).rbegin ( ) );
	}
}

int ExceptionHandler::handlerRec ( std::ostream & out, unsigned level, std::vector < std::function < int ( std::ostream &, unsigned ) > >::const_reverse_iterator iter ) {
	if ( iter == handlers ( ).rend ( ) ) {
		out << level << "unhandled exception." << std::endl;
		return 127;
	}
	try {
		return ( * iter ) ( out, level );
	} catch ( ... ) {
		return handlerRec ( out, level, std::next ( iter ) );
	}
}

int ExceptionHandler::handle ( std::ostream & out ) {
	return handlerRec ( out, 0, handlers ( ).rbegin ( ) );
}

} /* namespace alib */
